#include <time.h>
#include <zephyr/types.h>

#include "../includes/RTCHelper.h"
#include <zephyr/device.h>
#include <zephyr/drivers/rtc.h>
#include <zephyr/drivers/i2c.h>


//Defines for I2C communication
#define PCF8523_ALARM_ID 0

//Defines of device handles
static const struct device *boardRTCDevice = DEVICE_DT_GET(RTCNode);
static const struct i2c_dt_spec I2CDevice = I2C_DT_SPEC_GET(RTCNode);

//Status Struct
struct RTCState _RTC_internalState = {
    .timeValid = false,
    .clkOutEn = false
};

struct RTCState *boardRTCState = &_RTC_internalState;

//Time Get
int RTCgetTime(uint64_t *buffer){
    struct rtc_time rtcBuffer;
    int rslt = rtc_get_time( boardRTCDevice, &rtcBuffer);
    if (rslt != 0){  //Error Case
        _RTC_internalState.timeValid = false;
        return rslt;
    }
    *buffer = timeutil_timegm64(rtc_time_to_tm(&rtcBuffer));
    _RTC_internalState.timeValid = true;
    return 0;
}

//Time Set
int RTCsetTime(const uint64_t *timestamp){
    uint8_t capBits = 0x80;
    uint8_t clockout_disable = 0x38;
    i2c_burst_write_dt(&I2CDevice, 0x00, &capBits, 1);
    i2c_burst_write_dt(&I2CDevice, 0x0F, &clockout_disable, 1);
    return rtc_set_time( boardRTCDevice, (struct rtc_time*) gmtime(timestamp));
}

int RTCenableCLKout(){
    uint8_t regBuffer = 0x00;
    int rtrn = i2c_reg_read_byte_dt( &I2CDevice, 0x0F, &regBuffer);
    regBuffer = regBuffer & 0xF9;
    rtrn = i2c_reg_write_byte_dt( &I2CDevice, 0x0F, regBuffer);
    _RTC_internalState.clkOutEn = true;
    return rtrn;
}
int RTCdisableCLKout(){
    uint8_t regBuffer;
    int rtrn = i2c_burst_read_dt( &I2CDevice, 0x0F, &regBuffer, 1);
    regBuffer = regBuffer | 0x38;
    rtrn = i2c_burst_write_dt( &I2CDevice, 0x0F, &regBuffer, 1);
    _RTC_internalState.clkOutEn = false;
    return rtrn;
}

int RTCsetINTAlarm(uint64_t *alarmTime){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Alarm will be set to : %lld", (long long) *alarmTime);
    uint8_t buffer;
    uint16_t alarmMask;
    //Clear alarm flag in control register 2
    i2c_reg_read_byte_dt(&I2CDevice, 0x01, &buffer);
    i2c_reg_write_byte_dt(&I2CDevice, 0x01, (buffer & 0x77));
    //Set alarm
    rtc_alarm_get_supported_fields(boardRTCDevice, PCF8523_ALARM_ID, &alarmMask); //Return value not needed. By implementation it will always succed with ID = 0
    struct rtc_time *test = gmtime(alarmTime);
    int rtrn = rtc_alarm_set_time(boardRTCDevice, PCF8523_ALARM_ID, alarmMask, (struct rtc_time*) gmtime(alarmTime));
    rtrn = i2c_reg_read_byte_dt( &I2CDevice, 0x00, &buffer);
    buffer = buffer | 0x02;
    rtrn = i2c_reg_write_byte_dt( &I2CDevice, 0x00, buffer);
    //Add in a function to the shell that gets the current Alarm time from the device to allow for better testing. Need to get this issue fixed
    return rtrn;  
}

int RTCsetPSCountdown(int hours){
    uint8_t bFreqReg = 0x7F;
    uint8_t bEnableMask = 0x01;
    uint8_t bEnableInteruptMask = 0x01;
    uint8_t secondInteruptEnbMask = 0x04;
    int rtrn; 
    if (hours <= 0xFF){
        i2c_reg_write_byte_dt( &I2CDevice, 0x13, hours);
        rtrn = 0;
    }
    else {
        i2c_reg_write_byte_dt ( &I2CDevice, 0x13, 0xFF);
        rtrn = EDOM;
    } 
    i2c_reg_write_byte_dt( &I2CDevice, 0x12, bFreqReg); //Write frequency of the timer B clock
    uint8_t buffer;
    i2c_reg_read_byte_dt( &I2CDevice, 0x0F, &buffer);
    i2c_reg_write_byte_dt( &I2CDevice, 0x0F, (buffer | bEnableMask)); //Enable the b timer
    i2c_reg_read_byte_dt( &I2CDevice, 0x01, &buffer);
    i2c_reg_write_byte_dt( &I2CDevice, 0x01, (buffer | bEnableInteruptMask)); //Enable the B countdown interupt
    i2c_reg_read_byte_dt( &I2CDevice, 0x00, &buffer);
    i2c_reg_write_byte_dt( &I2CDevice, 0x00, (buffer | secondInteruptEnbMask)); //Enable the second interupt
    return rtrn;
}

void RTCclearTimerBInterupt(){
    uint8_t bClearMask = 0x20;
    uint8_t buffer;
    i2c_reg_read_byte_dt( &I2CDevice, 0x01, &buffer);
    i2c_reg_write_byte_dt( &I2CDevice, 0x01, (buffer & bClearMask));
}

//Console commands
int __shell_RTC_Get_Time(const struct shell *sh, size_t argc, char **argv){
    time_t buffer;
    RTCgetTime(&buffer);
    shell_print(sh, "Current Time in RTC %lld", (long long) buffer);
    return 0;
}

int __shell_RTC_Set_Time(const struct shell *sh, size_t argc, char **argv){
    const time_t setTime = atoll(argv[1]);
    return RTCsetTime(&setTime);
}

int __shell_RTC_EnableCLKOut(const struct shell *sh, size_t argc, char **argv){
    shell_print(sh, "Enabling the CLK output of the RTC");
    RTCenableCLKout();
    return 0;
}

int __shell_RTC_DisableCLKOut(const struct shell *sh, size_t argc, char **argv){
    shell_print(sh, "Disabling the CLK output of the RTC");
    RTCdisableCLKout();
    return 0;
}

int __shell_RTC_DUMP(const struct shell *sh, size_t argc, char **argv){
    uint8_t chip_memory[20];
    i2c_burst_read_dt(&I2CDevice, 0x00, chip_memory,20);
    shell_print(sh, "Printing values of battery charger chip");
    for (int i = 0; i < 20; i++){
        shell_print(sh, "Address: %x - Value: %x ", i, chip_memory[i] & 0xff);
    }
    return 0;
}

int __shell_RTC_alarmTest(const struct shell *sh, size_t argc, char **argv){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Setting alarm for 60 passed current time");
    uint64_t timeBuffer;
    RTCgetTime(&timeBuffer);
    timeBuffer += 60;
    RTCsetINTAlarm(&timeBuffer);
    LOG_INF("Alarm set");
    return 0;
}